import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="EELSFitter",
    version="0.0.5",
    author="A. Brokkelkamp, J.J. ter Hoeve, I. Postmes, L. Maduro, J. Rojo, S. Conesa Boj",
    author_email="j.j.ter.hoeve@vu.nl",
    description="Electron Energy Loss Spectroscopy Fitter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LHCfitNikhef/EELSfitter",
    project_urls={
        "Documentation": "https://lhcfitnikhef.github.io/EELSfitter",
    },
    classifiers=[
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        "ncempy>=1.8.0",
        "matplotlib>=3.3.2",
        "numpy>=1.19.2",
        "natsort>=7.1.0",
        "scipy>=1.5.2",
        "pandas>=1.0.5",
        "seaborn>=0.11.0",
        "torch>=1.6.0",
        "scikit_learn>=1.0",
        ],
    python_requires=">=3.7",
)



import numpy as np
import random
import os
import pandas as pd
import matplotlib
# matplotlib.use('Agg')
import matplotlib.pyplot as plt

import torch
import torch.nn as nn
import datetime as dt
import torch.optim as optim
from sklearn.model_selection import train_test_split
from matplotlib import rc


rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica'], 'size': 22})
rc('text', usetex=True)

from ..plotting.hyperparameters import plot_dE1



class MLP(nn.Module):
    """
    Multilayer Perceptron (MLP) class. It uses the following architecture

    .. math::

       [n_i, 10, 15, 5, n_f],

    where :math:`n_i` and :math:`n_f` denote the number of input features and output target values respectively.

    Parameters
    ----------
    num_inputs: int
        number of input features
    num_outputs: int
        dimension of the target output.
    """

    def __init__(self, num_inputs, num_outputs):
        super().__init__()
        # Initialize the modules we need to build the network
        self.linear1 = nn.Linear(num_inputs, 10)
        self.linear2 = nn.Linear(10, 15)
        self.linear3 = nn.Linear(15, 5)
        self.output = nn.Linear(5, num_outputs)
        self.sigmoid = nn.Sigmoid()
        self.relu = nn.ReLU()

    def forward(self, x):
        """
        Propagates the input features ``x`` through the MLP.
        Parameters
        ----------
        x: torch.tensor
            input features

        Returns
        -------
        x: torch.tensor
            MLP outcome
        """
        # Perform the calculation of the model to determine the prediction
        x = self.linear1(x)
        x = self.sigmoid(x)
        x = self.linear2(x)
        x = self.sigmoid(x)
        x = self.linear3(x)
        x = self.relu(x)
        x = self.output(x)
        return x


def scale(inp, ab):
    """
    Rescale the training data to lie between 0.1 and 0.9

    Parameters
    ----------
    inp: numpy.ndarray, shape=(M,)
        training data to be rescaled, e.g. :math:`\Delta E`
    ab: numpy.ndarray, shape=(M,)
        scaling parameters, which can be find with :py:meth:`find_scale_var() <find_scale_var>`.
    Returns
    -------
    Rescaled training data
    """
    return inp * ab[0] + ab[1]


def find_scale_var(inp, min_out=0.1, max_out=0.9):
    """
    Computes the scaling parameters needed to rescale the training data to lie between ``min_out`` and ``max_out``.

    Parameters
    ----------
    inp: numpy.ndarray, shape=(M,)
        training data to be rescaled
    min_out: float
        lower limit. Set to 0.1 by default.
    max_out: float
        upper limit. Set to 0.9 by default

    Returns
    -------
    a, b: list
        list of rescaling parameters
    """
    a = (max_out - min_out) / (inp.max() - inp.min())
    b = min_out - a * inp.min()
    return [a, b]


def weight_reset(m):
    """
    Reset the weights and biases associated with the model ``m``.

    Parameters
    ----------
    m: MLP
        Model of type :py:meth:`MLP <MLP>`.
    """
    if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
        m.reset_parameters()


def loss_fn(output, target, error):
    r"""
    The loss function to train the ZLP takes the model ``output``, the raw spectrum ``target`` and
    the associated ``error``. The latter corresponds to the one sigma spread within a given cluster
    at fixed :math:`\Delta E`. It returns the cost function :math:`C_{\mathrm{ZLP}}^{(m)}` associated with the replica :math:`m` as

    .. math::

       C_{\mathrm{ZLP}}^{(m)} = \frac{1}{n_{E}}\sum_{k=1}^K \sum_{\ell_k=1}^{n_E^{(k)}} \frac{\left[ I^{(i_{m,k},
       j_{m,k})}(E_{\ell_k}) - I_{\rm ZLP}^{({\mathrm{ NN}})(m)} \left(E_{\ell_k},\ln \left( N_{\mathrm{ tot}}^{(i_{
       m,k},j_{m,k})}\right) \right) \right]^2}{\sigma^2_k \left( E_{\ell_k}\right) }.


    Parameters
    ----------
    output: torch.tensor
        Neural Network output
    target: torch.tensor
        Raw EELS spectrum
    error: torch.tensor
        Uncertainty on :math:`\log I_{\mathrm{EELS}}(\Delta E)`.

    Returns
    -------
    loss: torch.tensor
        Loss associated with the model ``output``.
    """
    loss = torch.mean(torch.square((output - target) / error))
    return loss


def smooth(data, window_len=10, window='hanning', keep_original=False):
    """
    smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.
    
    Parameters
    ----------
    data: numpy.ndarray, shape=(M,)
        The input signal
    window_len: int, optional
        The dimension of the smoothing window; should be an odd integer. Set to 10 by default.
    window: str, optional
        the type of window from ``"flat"``, '``"hanning"``, ``"hamming"``, ``"bartlett"``, ``"blackman"``
        ``"flat"`` window will produce a moving average smoothing.

    Returns
    -------
    smoothed_signal: numpy.ndarray, shape=(M,)
        The smoothed spectrum

    """
    window_len += (window_len + 1) % 2
    s = np.r_['-1', data[:, window_len - 1:0:-1], data, data[:, -2:-window_len - 1:-1]]

    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.' + window + '(window_len)')

    surplus_data = int((window_len - 1) * 0.5)
    smoothed_signal = np.apply_along_axis(lambda m: np.convolve(m, w / w.sum(), mode='valid'), axis=1, arr=s)[:,
           surplus_data:-surplus_data]
    return smoothed_signal


def smooth_clusters(image, clusters, window_len=None):
    """
    Smooth the spectrum for all clusters

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <spectral_image.SpectralImage>` object
    clusters: numpy.ndarray, shape=(M,)

    window_len: int
        Smoothing window length

    Returns
    -------
    smoothed_clusters: numpy.ndarray, shape=(M,)
        The smoothed spectra for all clusters
    """
    smoothed_clusters = np.zeros((len(clusters)), dtype=object)
    for i in range(len(clusters)):
        smoothed_clusters[i] = smooth(clusters[i])
    return smoothed_clusters


def derivative_clusters(image, clusters):
    """
    Computes the derivative of the spectral image in each cluster.

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <spectral_image.SpectralImage>` object
    clusters: numpy.ndarray, shape=(M,)
        An array with size the number of clusters Each entry is a 2D array that contains all the spectra within the cluster.

    Returns
    -------
    der_clusters: numpy.ndarray, shape=(M,)
        Slope of the intensity spectrum for all clusters.
    """
    dx = image.ddeltaE
    der_clusters = np.zeros((len(clusters)), dtype=object)
    for i in range(len(clusters)):
        der_clusters[i] = (clusters[i][:, 1:] - clusters[i][:, :-1]) / dx
    return der_clusters


def find_min_de1(image, dy_dx, y_smooth):
    """
    Finds the location of the first local minimum of the intensity spectrum

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <EELSFitter.core.spectral_image.SpectralImage>` object
    dy_dx: numpy.ndarray, shape=(M,)
        Derivative of the intensity spectrum
    y_smooth: numpy.ndarray, shape=(M,)
        Smoothed intensity spectrum

    Returns
    -------
    pos_der: float
        Position of first local minimum
    """
    crossing = (dy_dx > 0)
    if not crossing.any():
        print("shouldn't get here")
        up = np.argmin(np.absolute(dy_dx)[np.argmax(y_smooth) + 1:]) + np.argmax(y_smooth) + 1
    else:
        up = np.argmax(crossing[np.argmax(y_smooth) + 1:]) + np.argmax(y_smooth) + 1
    pos_der = image.deltaE[up]
    return pos_der


def determine_de1(image, dy_dx_clusters, y_smooth_clusters, shift_de1=0.7, perc_de1=16, plot_de1=False, prob=True):
    """
    Computes the hyperparamter :math:`\Delta E_I` for every cluster in two ways:

    - Take a certain fraction of the location of the first local minimum. The fraction is tuned by ``shift_de1``.
    - Take :math:`\Delta E_I` such that only 16% of the replicas have a positive slope.

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <spectral_image.SpectralImage>` object.
    dy_dx_clusters: numpy.ndarray, shape=(M,)
        An array that contains an array for each cluster, which subsequently contains the slope of the spectrum at each
        pixel within the cluster.
    y_smooth_clusters: numpy.ndarray, shape=(M,)
        An array that contains an array for each cluster, which subsequently contains the smoothed spectrum at each
        pixel within the cluster.
    shift_de1: float, optional
        Shift the location of :math:`\Delta E_I` by a factor of ``shift_dE1`` w.r.t. to the first local minimum.
    perc_de1: float, optional
        Percentage of replicas that is allowed to have a positive slope at :math:`\Delta E_I`.
    plot_de1: bool, optional
        When set to `True`, the location of :math:`\Delta E_I` is plotted. Set to `False` by default.
    prob: bool, optional
        When set to `True`, the location of :math:`\Delta E_I` is determined by demanding that ``perc_de1``% of the EELS replicas should
        have a positive slope.

    Returns
    -------
    dE1_clusters: numpy.ndarray, shape=(M,)
        Array with the value of :math:`\Delta E_I` for each cluster.
    """

    # number of clusters
    n_clusters = len(y_smooth_clusters)

    # median EELS spectrum for each cluster
    y_smooth_clusters_avg = np.array([np.average(y_smooth_clusters[i], axis=0) for i in range(n_clusters)])

    # average slope of the EElS spectrum for each cluster
    dy_dx_avg = np.array([np.average(dy_dx_clusters[i], axis=0) for i in range(n_clusters)])

    # zeros of the first derivative for each cluster
    min_clusters = np.array([find_min_de1(image, dy_dx_avg[i], y_smooth_clusters_avg[i]) for i in range(n_clusters)])

    # find the replica such that 16% of the replicas have a steeper slope
    cl_high = 100 - perc_de1
    dy_dx_perc = np.array([np.nanpercentile(dy_dx_clusters[i], cl_high, axis=0) for i in range(n_clusters)])

    # find the index at which 16% of the replicas have a positive slope
    # and evaluate the corresponding value of deltaE.

    #TODO: not consistent: gives extra solutions sometimes depending on the clustering 29/07/21
    idx = np.argwhere(np.diff(np.sign(dy_dx_perc)))[:, 1][1::2]
    de1_prob = image.deltaE[idx]

    # find dE1 by taking a certain fraction of the location of the first local minimum.
    # The fraction is controlled by shift_dE1
    de1_shift = np.array([min_clusters[i] * shift_de1 for i in range(n_clusters)])

    # number of replicas per cluster
    n_rep_cluster = np.array([dy_dx_clusters[i].shape[0] for i in range(n_clusters)])

    # display the computed values of dE1 (both methods) together with the raw EELS spectrum
    if plot_de1:
        print("Plotting choice of dE1...")
        fig_1, fig_2 = plot_dE1(image, y_smooth_clusters, dy_dx_clusters, min_clusters, de1_prob, de1_shift, prob)
        fig_1.savefig(os.path.join(image.output_path, "der.pdf"))
        fig_2.savefig(os.path.join(image.output_path, "raw.pdf"))
        print("Plots stored in {}".format(image.output_path))
    # return the de1 selected by the bool prob
    if prob:
        return de1_prob
    else:
        return de1_shift


def filter_bad_bins(image, spectra):
    #TODO: add dimensions to spectra in docstring
    """
    Filter out experimental hiccups.

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <spectral_image.SpectralImage>` object.
    spectra:

    Returns
    -------
    output:
        Spectral image with experimental hiccups removed.
    """
    spectra_filtered = np.zeros(len(spectra), dtype=object)
    for j, spectra_cluster in enumerate(spectra):
        spectra_filtered_clustered = []
        for i, spectra_cluster_model in enumerate(spectra_cluster):
            hiccup_check = np.max(spectra_cluster_model[(image.deltaE > 1.0) & (image.deltaE < 3.0)])
            if hiccup_check > 2500:
                continue
            else:
                spectra_filtered_clustered.append(spectra_cluster_model)
        spectra_filtered_clustered = np.array(spectra_filtered_clustered)
        spectra_filtered[j] = spectra_filtered_clustered
    return spectra_filtered


def train_zlp_scaled(image, spectra, n_rep=500, n_epochs=30000, lr=1e-3, shift_dE1=0.7, shift_dE2=3,
                     path_to_models="models",
                     display_step=1000, bs_rep_num=0, perc_de1=16, plot_de1=False, prob=True):
    """
    Trains the ZLP on the (clustered) spectral image.

    One should train on two input features: the (scaled) log intensity :math:`\log I_{\mathrm{EELS}}` and the (scaled) :math:`\Delta E`.

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <spectral_image.SpectralImage>` object.
    spectra: numpy.ndarray, shape=(M,)
        An array of size ``n_clusters``, with each entry being a 2D array that contains all the spectra within the cluster.
    n_rep: int, optional
        Number of replicas to train on. Each replica consists of one spectrum from each cluster. Set to 500 by default.
    n_epochs: int, optional
        Number of epochs. Set to 30000 by default
    lr: float, optional
        Learning rate. Set to :math:`10^{-3}` by default.
    shift_dE1: float, optional
        Set to 0.7 by default. When given, :math:`\Delta E_I` is located at ``shift_dE1`` times the first local minimum in the spectrum.
    path_to_models: str, optional
        Path to where the trained models should be stored.
    display_step: int, optional
        Number of epochs after which to print output.
    bs_rep_num: int, optional
        For cluster users only: labels the parallel run.
    perc_de1: float, optional
        Percentage of replicas that is allowed to have a positive slope at :math:`\Delta E_I`.
    plot_de1: bool, optional
        When set to `True`, the location of :math:`\Delta E_I` is plotted. Set to `False` by default.
    prob: bool, optional
        When set to `True`, the location of :math:`\Delta E_I` is determined by demanding that 16% of the EELS replicas should
        have a positive slope.

    Examples
    --------
    To train the ZLP, one can run the following lines of code:

    >>> dm4_path = 'path to dm4 file'
    >>> im = SpectralImage.load_data(dm4_path)
    >>> training_data = im.get_cluster_spectra(conf_interval=1, signal='EELS')
    >>> train_zlp_scaled(im, training_data, n_clusters=5, n_rep=500, n_epochs=1000, path_to_models='path', display_step=10)
    """

    if display_step is None:
        print_progress = False
        display_step = 1E6
    else:
        print_progress = True

    if not os.path.exists(path_to_models):
        os.mkdir(path_to_models)

    # set all intensities smaller than 1 to 1
    for i in range(image.n_clusters):
        spectra[i][spectra[i] < 1] = 1

    loss_test_reps = np.zeros(n_rep)
    loss_train_reps = np.zeros(n_rep)

    sigma_clusters = np.zeros((image.n_clusters, image.l))  # shape = (n_clusters, n dE)
    for cluster in range(image.n_clusters):
        ci_low = np.nanpercentile(np.log(spectra[cluster]), 16, axis=0)
        ci_high = np.nanpercentile(np.log(spectra[cluster]), 84, axis=0)
        sigma_clusters[cluster, :] = np.absolute(ci_high - ci_low)

    wl1 = round(image.l / 20)
    wl2 = wl1 * 2

    # filter out bad bins
    #spectra_filtered = filter_bad_bins(image, spectra)

    # arrays with image.n_clusters sub-arrays, each having shape (n_spectra in cluster, n dE)
    #spectra_smooth = smooth_clusters(image, spectra_filtered, wl1)
    spectra_smooth = smooth_clusters(image, spectra, wl1)
    dy_dx = derivative_clusters(image, spectra_smooth)
    smooth_dy_dx = smooth_clusters(image, dy_dx, wl2)

    # hyperparameters: discard training data > dE1 and add pseudo data after dE2
    dE1 = determine_de1(image, smooth_dy_dx, spectra_smooth, shift_dE1, perc_de1, plot_de1, prob)
    dE2 = shift_dE2 * dE1

    if print_progress:
        print("dE1 & dE2:", np.round(dE1, 3), dE2)

    # rescale the dE features to [0.1, 0.9]
    ab_deltaE = find_scale_var(image.deltaE)
    deltaE_scaled = scale(image.deltaE, ab_deltaE)

    # collect all spectra from all clusters into a single array
    all_spectra = np.empty((0, image.l))
    for i in range(len(spectra)):
        all_spectra = np.append(all_spectra, spectra[i], axis=0)

    # the total integrated intensity for all pixels and rescale
    int_log_I = np.log(np.sum(all_spectra, axis=1)).flatten()
    ab_int_log_I = find_scale_var(int_log_I)
    del all_spectra

    path_scale_var = os.path.join(path_to_models, "scale_var.txt")
    if not os.path.exists(path_scale_var):
        np.savetxt(path_scale_var, ab_int_log_I)

    path_dE1 = os.path.join(path_to_models, "dE1.txt")
    if not os.path.exists(path_dE1):
        np.savetxt(path_dE1, np.vstack((image.clusters, dE1)))

    # path to file that keeps track of the cost function after training for each replica
    train_cost_path = os.path.join(path_to_models, "costs_train_{}.txt".format(bs_rep_num))
    test_cost_path = os.path.join(path_to_models, "costs_test_{}.txt".format(bs_rep_num))

    for i in range(n_rep):
        save_idx = i + n_rep * bs_rep_num - n_rep + 1
        nn_rep_path = os.path.join(path_to_models, "nn_rep_{}".format(save_idx))

        training_report_path = os.path.join(path_to_models, 'rep_{}'.format(save_idx))
        if not os.path.exists(training_report_path):
            os.makedirs(training_report_path)

        if print_progress: print("Started training on replica number {}".format(i) + ", at time ", dt.datetime.now())
        data_y = np.empty((0, 1))
        data_x = np.empty((0, 2))
        data_sigma = np.empty((0, 1))

        for cluster in range(image.n_clusters):
            n_cluster = len(spectra[cluster])  # number of spectra in a cluster

            # select a random spectrum from the cluster
            idx = random.randint(0, n_cluster - 1)

            # create slicing indices for the regions I, II and III
            select1 = len(image.deltaE[image.deltaE < dE1[cluster]])
            select2 = len(image.deltaE[image.deltaE > dE2[cluster]])

            # Only use the log intensity up to dE1 as data. Do this for a random spectrum for each cluster.
            data_y = np.append(data_y, np.log(spectra[cluster][idx][:select1]))
            # The log intensity is set to zero in region III
            data_y = np.append(data_y, np.zeros(select2))

            # pseudo_x contains the two input features: log integrated I and deltaE
            features = np.ones((select1 + select2, 2))
            features[:select1, 0] = deltaE_scaled[:select1]
            features[-select2:, 0] = deltaE_scaled[-select2:]

            # rescale the integrated intensity
            int_log_I_idx_scaled = scale(np.log(np.sum(spectra[cluster][idx])), ab_int_log_I)
            features[:, 1] = int_log_I_idx_scaled

            data_x = np.concatenate((data_x, features))
            data_sigma = np.append(data_sigma, sigma_clusters[cluster][:select1])
            data_sigma = np.append(data_sigma, 0.8 * np.ones(select2))

        model = MLP(num_inputs=2, num_outputs=1)
        model.apply(weight_reset)
        optimizer = optim.Adam(model.parameters(), lr=lr)

        train_x, test_x, train_y, test_y, train_sigma, test_sigma = train_test_split(data_x, data_y, data_sigma,
                                                                                     test_size=0.4)

        n_test = len(test_x)
        n_train = len(train_x)

        test_x = test_x.reshape(n_test, 2)
        test_y = test_y.reshape(n_test, 1)
        train_x = train_x.reshape(n_train, 2)
        train_y = train_y.reshape(n_train, 1)
        train_sigma = train_sigma.reshape(n_train, 1)
        test_sigma = test_sigma.reshape(n_test, 1)

        train_x = torch.from_numpy(train_x)
        train_y = torch.from_numpy(train_y)
        train_sigma = torch.from_numpy(train_sigma)
        test_x = torch.from_numpy(test_x)
        test_y = torch.from_numpy(test_y)
        test_sigma = torch.from_numpy(test_sigma)

        # store the test and train loss per epoch
        loss_test_n = []
        loss_train_n = []

        n_stagnant = 0
        n_stagnant_max = 5
        for epoch in range(1, n_epochs + 1):

            # set the model to training mode
            model.train()
            output = model(train_x.float())
            loss_train = loss_fn(output, train_y, train_sigma)
            loss_train_n.append(loss_train.item())

            optimizer.zero_grad()
            loss_train.backward()
            optimizer.step()

            # set the model to evaluation mode
            model.eval()
            with torch.no_grad():
                output_test = model(test_x.float())
                loss_test = loss_fn(output_test, test_y, test_sigma)
                loss_test_n.append(loss_test.item())
                if epoch % display_step == 0 and print_progress:
                    print('Rep {}, Epoch {}, Training loss {}, Testing loss {}'.format(i, epoch,
                                                                                       round(loss_train.item(), 3),
                                                                                       round(loss_test_n[epoch - 1], 3)))

                # update the test and train loss of the replica
                loss_test_reps[i] = loss_test.item()
                loss_train_reps[i] = loss_train.item()

        # save the model state when the maximum number of allowed epochs has been reached
        torch.save(model.state_dict(), nn_rep_path)

        # make a training report for each replica
        training_report(training_report_path, loss_train_n, loss_test_n)

    with open(train_cost_path, "w") as text_file:
        for item in loss_train_reps:
            text_file.write("%s\n" % item)

    with open(test_cost_path, "w") as text_file:
        for item in loss_test_reps:
            text_file.write("%s\n" % item)


def training_report(path, loss_train, loss_test):
    """
    Produce a training report: evolution of the training and validation loss per epoch

    Parameters
    ----------
    path: str
        location where to store the evolution plots
    loss_train: numpy.ndarray, shape=(M,)
        Training loss
    loss_test: numpy.ndarray, shape=(M,)
        Validation loss
    """
    training_loss_path = os.path.join(path, 'training_loss.txt')
    with open(training_loss_path, 'w') as f:
        for item in loss_train:
            f.write("%s\n" % item)

    test_loss_path = os.path.join(path, 'test_loss.txt')
    with open(test_loss_path, 'w') as f:
        for item in loss_test:
            f.write("%s\n" % item)

    fig, ax = plt.subplots(figsize=(1.1*10, 1.1*6))
    plt.plot(loss_train, label=r"$\rm{Training\;loss}$")
    plt.plot(loss_test, label=r"$\rm{Validation\;loss}$")
    plt.xlabel(r"$\rm{Epochs}$")
    plt.ylabel(r"$\rm{Loss}$")
    plt.legend(loc='best', frameon=False)
    fig.savefig(os.path.join(path, 'loss.pdf'))


def ewd(x, nbins):
    #TODO: what is the purpose of this function?
    """
    Apply Equal Width Discretization (EWD) to ``x`` to determine variances

    Parameters
    ----------
    x: numpy.ndarray, shape=(M,)
        input data
    nbins: int
        number of bins

    Returns
    -------
    cuts1: numpy.ndarray, shape=(M,)
    cuts2: numpy.ndarray, shape=(M,)


    """

    cuts1, cuts2 = pd.cut(x, nbins, retbins=True)

    return cuts1, cuts2


def CI_high(data, confidence=0.68):
    """
    Computes the upper 68% confidence level (by default) of ``data``.

    Parameters
    ----------
    data: numpy.ndarray, shape=(M,)
        arbitrary array
    confidence: float
        Confidence level

    Returns
    -------
    high_a: float
        confidence level set by ``confidence``.
    """

    a = np.array(data)
    n = len(a)
    b = np.sort(data)

    highest = np.int((1 - (1 - confidence) / 2) * n)
    high_a = b[highest]

    return high_a


def CI_low(data, confidence=0.68):
    """
    Computes the lower 68% confidence level (by default) of ``data``.

    Parameters
    ----------
    data: numpy.ndarray, shape=(M,)
        arbitrary array
    confidence: float
        Confidence level

    Returns
    -------
    low_a: float
        confidence level set by ``confidence``.
    """

    a = np.array(data)
    n = len(a)
    b = np.sort(data)
    lowest = np.int(((1 - confidence) / 2) * n)
    low_a = b[lowest]

    return low_a

def get_mean(data):
    #TODO: trivial function, remove
    """
    Returns the mean of ``data``.

    Parameters
    ----------
    data: numpy.ndarray, shape=(M,)
        data to average
    Returns
    -------
    mean: numpy.ndarray, shape=(M,)
        The mean of ``data``.
    """
    mean = np.mean(data)
    return mean



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 18 16:13:47 2021

@author: isabel
"""
import numpy as np
#import sys
#import os
#import pickle
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib import rc

from scipy.optimize import curve_fit
from scipy.signal import savgol_filter
from spectral_image import SpectralImage

#%%

# path_to_results = "C:/Users/abelbrokkelkam/PhD/data/MLdata/results/dE_n10-inse_SI-003/E1_shift09_k10_average/fixed_b/image_KK_4.pkl"
path_to_results = "C:/Users/abelbrokkelkam/PhD/data/MLdata/results/dE_nf-ws2_SI-001/E1_p16_k5_average/fixed_b/image_KK.pkl"
im = SpectralImage.load_spectral_image(path_to_results)

# path_to_image = 'C:/Users/abelbrokkelkam/PhD/data/dmfiles/10n-dop-inse-B1_stem-eels-SI-processed_003.dm4'
# path_to_image = 'C:/Users/abelbrokkelkam/PhD/data/dmfiles/area03-eels-SI-aligned.dm4'
# im = SpectralImage.load_data(path_to_image)

#path_to_models = 'C:/Users/abelbrokkelkam/PhD/data/MLdata/models/dE_nf-ws2_SI-001/E1_p16_k5_average/'
#im.load_zlp_models(path_to_models=path_to_models)
im.pool(5)
im.cluster(5)
im.calc_axes()

#%% Settings for all heatmaps
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica'], 'size': 12})
rc('text', usetex=True)

#%%

# InSe general settings
# cmap="coolwarm"
# npix_xtick=24.5
# npix_ytick=12.25
# sig_ticks = 3
# scale_ticks = 1E-3
# tick_int = True
# thicknesslimit = np.nanpercentile(im.t[im.clustered == 0],0)
# mask = im.t[:,:,0] < thicknesslimit
# cb_scale=0.4
# title_specimen = r'$\rm{InSe\;}$'
# save_title_specimen = 'InSe_flake'
# #save_loc = "C:/Users/abelbrokkelkam/PhD/data/MLdata/plots/dE_n10-inse_SI-003/E1_shift09_k10_average/fixed_b/pdfplots/"
# save_loc = "C:/Users/abelbrokkelkam/PhD/data/MLdata/plots/paper/InSe/indirect/"

# im.e0 = 200									# keV
# im.beta = 21.3								# mrad
# im.set_n(3.0)								# refractive index, InSe no background

# WS2 general settings
cmap="coolwarm"
npix_xtick=26.25
npix_ytick=26.25
sig_ticks = 3
scale_ticks = 1E-3
tick_int = True
thicknesslimit = np.nanpercentile(im.t[im.clustered == 2],99)
mask = ((np.isnan([im.t[:,:,0]])[0]) | (im.t[:,:,0] > thicknesslimit))
cb_scale=0.85
title_specimen = r'$\rm{WS_2\;nanoflower\;}$'
save_title_specimen = 'WS2_nanoflower_flake'
# save_loc = "C:/Users/abelbrokkelkam/PhD/data/MLdata/plots/dE_nf-ws2_SI-001//E1_p16_k5_average/fixed_b/pdfplots/"
save_loc = "C:/Users/abelbrokkelkam/PhD/data/MLdata/plots/paper/WS2/"

im.e0       = 200                           # keV
im.beta     = 67.2                          # mrad
im.set_n(4.1462, n_background = 2.1759)     # refractive index, WS2 triangles with SiN substrate as background
im.set_n(4.1462, n_background = 1)          # refractive index, WS2 nanoflower with vacuum as background, note that calculations on SiN may get weird

def round_to_nearest(value, base=5):
    return base * round(float(value) / base)


#%% CLUSTER
im.plot_heatmap(im.clustered, title = title_specimen + r'$\rm{-\;K=10\;clusters\;}$', 
                cbar_kws={'label': r'$\rm{[-]\;}$','shrink':cb_scale}, discrete_colormap = True,
                xlab = r'$\rm{[nm]\;}$', ylab = r'$\rm{[nm]\;}$', cmap = cmap,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Clustered')

#%% THICKNESS
im.plot_heatmap(im.t[:,:,0], title = title_specimen + r"$\rm{-\;Thickness\;}$", 
                cbar_kws={'label': r"$\rm{[nm]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask, vmin=0,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Thickness')

im.plot_heatmap(im.t[:,:,0], title = title_specimen + r"$\rm{-\;Thickness\;}$", 
                cbar_kws={'label': r"$\rm{[nm]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap, 
                mask = mask, vmin = 0, vmax = 60, 
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Thickness_capped')

im.plot_heatmap((im.t[:,:,2]-im.t[:,:,1])/(2*im.t[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Thickness\;}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask, vmin = 0,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Thickness_Error')

im.plot_heatmap((im.t[:,:,2]-im.t[:,:,1])/(2*im.t[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Thickness\;}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask, vmin = 0, vmax = 0.025,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Thickness_Error_capped')

# im.plot_heatmap((im.t[:,:,2]-im.t[:,:,1])/(im.t[:,:,0]), title = title_specimen + " - r"$\rm{-\;Relative\;Broadness\;CI\;Thickness\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmin = 0, vmax = 0.02,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Thickness_CI')
#%% THICKNESS DISCRETIZED

# mask_t = (mask | ((im.t[:,:,2]-im.t[:,:,1])/im.t[:,:,0] >= 1))
# size_t_bins = np.nanpercentile((im.t[:,:,2]-im.t[:,:,1])[~mask_t],100)/0.3
# t_round  = np.round(im.t[:,:,0]/size_t_bins) * size_t_bins
# im.plot_heatmap(t_round, title = title_specimen + r"$\rm{-\;Thickness\;Discretized\;}$",
#                 cbar_kws={'label': r"$\rm{[nm]\;}$",'shrink':0.4}, color_bin_size = size_t_bins, discrete_colormap = True,
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask_t, vmax = 300, vmin = 50,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_thickness_Discretized')

#%% MAX IEELS

im.plot_heatmap(im.max_ieels[:,:,0], title = title_specimen + r"$\rm{-\;Maximum\;IEELS\;}$", 
                cbar_kws={'label': 'Energy loss [eV]','shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Max_IEELS')

im.plot_heatmap((im.max_ieels[:,:,2]-im.max_ieels[:,:,1])/(2*im.max_ieels[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Maximum\;IEELS\;}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask, vmax = 0.001, 
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Max_IEELS_Error')

# im.plot_heatmap((im.max_ieels[:,:,2]-im.max_ieels[:,:,1])/(im.max_ieels[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Broadness\;CI\;Maximum\;IEELS\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 0.001,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Max_IEELS_CI')

#%% MAX IEELS DISCRETIZED

mask_max_ieels = (mask | ((im.max_ieels[:,:,2]-im.max_ieels[:,:,1])/im.max_ieels[:,:,0] >= 1))
size_ieels_bins = 1.0 # round_to_nearest(np.nanpercentile((im.max_ieels[:,:,0])[~mask_max_ieels],50)/2,0.5)
ieels_round  = np.round(im.max_ieels[:,:,0]/size_ieels_bins) * size_ieels_bins
im.plot_heatmap(ieels_round, title = title_specimen + r"$\rm{-\;Maximum\;IEELS\;Discretized\;}$", 
                cbar_kws={'label': r"$\rm{Energy\;Loss\;[eV]\;}$", 'shrink':cb_scale}, color_bin_size = size_ieels_bins, discrete_colormap = True,
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask, vmin = 21, vmax = 25,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Max_IEELS_Discretized')


#%% THICKNESS CROSSSECTION

fig1, ax1 = plt.subplots(dpi=200)
ax1.set_title(title_specimen + r"$\rm{-\;Thickness\;y\;cross\;section}$")
ax1.set_xlabel(r"$\rm{x-axis\;[nm]\;}$")
ax1.set_ylabel(r"$\rm{Thickness\;[nm]\;}$")
for i in np.arange(5,len(im.y_axis),5):
    row = i
    colors = cm.coolwarm(np.linspace(0,1,len(im.t[0,:,0])))
    ax1.set_prop_cycle(color=colors)
    for j in range(len(im.t[row,:,0]) - 1):
        ax1.plot(im.x_axis[j:j + 2]*scale_ticks, im.t[row,:,0][j:j + 2])
        ax1.fill_between(im.x_axis[j:j + 2]*scale_ticks, im.t[row,:,2][j:j + 2], im.t[row,:,1][j:j + 2], alpha = 0.3)
    
    
#     ax1.plot(im.x_axis, im.t[row,:,0], label = "Row " + str(row))
#     ax1.fill_between(im.x_axis, im.t[row,:,2], im.t[row,:,1], alpha = 0.3)
# ax1.get_xaxis().set_major_formatter(plt.FuncFormatter(lambda x, p: int(x*scale_ticks)))
# ax1.legend(loc=2)


fig2, ax2 = plt.subplots(dpi=200)
ax2.set_title(title_specimen + r"$\rm{-\;Thickness\;x\;cross\;section}$")
ax2.set_xlabel(r"$\rm{y-axis\;[nm]\;}$")
ax2.set_ylabel(r"$\rm{Thickness\;[nm]\;}$")
for i in np.arange(5,len(im.x_axis),5):
    column = i
    colors = cm.coolwarm(np.linspace(0,1,len(im.t[0,:,0])))
    ax2.plot(im.y_axis*scale_ticks, im.t[:,column,0], color = colors[i])
    # ax2.plot(im.y_axis, im.t[:,column,0], label = "Column " + str(column), color = colors[i])
    ax2.fill_between(im.y_axis*scale_ticks, im.t[:,column,2], im.t[:,column,1], alpha = 0.3, color = colors[i])
# ax2.get_xaxis().set_major_formatter(plt.FuncFormatter(lambda x, p: int(x*scale_ticks)))
# ax2.legend(loc=2)


#%% NUM CROSSINGS

im.n_cross = np.round(im.n_cross)

mask_cross = (mask | (im.n_cross[:,:,0] == 0))

im.plot_heatmap(im.n_cross[:,:,0], title = title_specimen + r"$\rm{-\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Nr.\;Crossings\;}$",'shrink':cb_scale}, discrete_colormap = True,
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Crossings')

im.plot_heatmap((im.n_cross[:,:,2]-im.n_cross[:,:,1]), title = title_specimen + r"$\rm{-\;Relative\;Broadness\;CI\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Nr.\;Crossings\;}$",'shrink':cb_scale}, discrete_colormap = True, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Crossings_CI')

# mask_cross = (mask | (im.n_cross[:,:,2] == 0))
# im.plot_heatmap(im.n_cross[:,:,0], title = "Indium Selenide Sample \nCrossings $\u03B5_{1}$ \nAlternative mask", cbar_kws={'label': 'nr. crossings','shrink':0.4}, cmap = cmap, mask = mask_cross, discrete_colormap = True, n_xticks=8, n_yticks=6)
# im.plot_heatmap((im.n_cross[:,:,2]-im.n_cross[:,:,1]), title = "Indium Selenide Sample \nRelative broadness CI Crossings $\u03B5_{1}$ \nAlternative mask", cbar_kws={'label': 'nr. crossings','shrink':0.4}, cmap = cmap, mask = mask_cross, discrete_colormap = True, n_xticks=8, n_yticks=6)


#%% ENERGY AT FIRST CROSSINGS

first_crossings = np.zeros(np.append(im.image_shape, 3))
first_crossings_CI = np.zeros(im.image_shape)
for i in range(im.image_shape[0]):
    for j in range(im.image_shape[1]):
        if type(im.E_cross[i,j]) == np.ndarray:
            if len(im.E_cross[i,j]) > 0:
                first_crossings[i,j,:] = im.E_cross[i,j][0,:]
                first_crossings_CI[i,j] = (im.E_cross[i,j][0,2]-im.E_cross[i,j][0,1])/(2*im.E_cross[i,j][0,0])
        
mask_cross = (mask | (im.n_cross[:,:,0] == 0))        
im.plot_heatmap(first_crossings[:,:,0], title = title_specimen + r"$\rm{-\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross, 
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_Crossings')

im.plot_heatmap(first_crossings_CI, title = title_specimen + r"$\rm{-\;Relative\;Error\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_First_Crossings_Error')

im.plot_heatmap(first_crossings_CI, title = title_specimen + r"$\rm{-\;Relative\;Error\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross, vmax = 0.01,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_First_Crossings_Error_capped')

# im.plot_heatmap(first_crossings_CI, title = title_specimen + r"$\rm{-\;Relative\;Broadness\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask_cross, vmax = 0.01,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Energy_First_Crossings_CI')

# mask_cross = (mask | (im.n_cross[:,:,2] == 0))
# im.plot_heatmap(first_crossings[:,:,0], title = "Indium Selenide Sample \nenergy first crossings $\u03B5_{1}$ \n(for chance at least 1 crossing > 0.1) \nAlternative mask", cbar_kws={'label': 'energy [eV]','shrink':0.4}, cmap = cmap, mask = mask_cross, n_xticks=8, n_yticks=6)
# im.plot_heatmap(first_crossings_CI, title = "Indium Selenide Sample \nrelative broadness CI energy first crossings $\u03B5_{1}$ \n(for chance at least 1 crossing > 0.1) \nAlternative mask", cbar_kws={'label': 'ratio [-]','shrink':0.4}, cmap = cmap, mask = mask_cross, n_xticks=8, n_yticks=6)
# im.plot_heatmap(first_crossings_CI, title = "Indium Selenide Sample \nrelative broadness CI energy first crossings $\u03B5_{1}$ \n(for chance at least 1 crossing > 0.1), capped at 0.005 \nAlternative mask", cbar_kws={'label': 'ratio [-]','shrink':0.4}, vmax = 0.005, cmap = cmap, mask = mask_cross, n_xticks=8, n_yticks=6)

#%% CROSSINGS AT MAX IEELS

mask_max_cross = (mask | (im.n_cross[:,:,0] == 0) | (first_crossings[:,:,0] < 21) | (first_crossings[:,:,0] > 25) )        
im.plot_heatmap(first_crossings[:,:,0], title = title_specimen + r"$\rm{-\;Energy\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_max_cross, vmin = 21, vmax = 25, 
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_Max_Crossings')

im.plot_heatmap(first_crossings_CI, title = title_specimen + r"$\rm{-\;Relative\;Error\;Energy\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, 
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_max_cross, vmax = 0.01,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_Max_Crossings_Error')

# im.plot_heatmap(first_crossings_CI, title = title_specimen + r"$\rm{-\;Relative\;Broadness\;Energy\;Crossings\;}$" + "$\epsilon_{1}$" + r"$\rm{\;IEELS\;Max\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask_cross, vmax = 0.01,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Energy_Max_Crossings_CI')

# mask_cross = (mask | (im.n_cross[:,:,2] == 0))
# im.plot_heatmap(first_crossings[:,:,0], title = "Indium Selenide Sample \nenergy first crossings $\u03B5_{1}$ \n(for chance at least 1 crossing > 0.1) \nAlternative mask", cbar_kws={'label': 'energy [eV]','shrink':0.4}, cmap = cmap, mask = mask_cross, n_xticks=8, n_yticks=6)
# im.plot_heatmap(first_crossings_CI, title = "Indium Selenide Sample \nrelative broadness CI energy first crossings $\u03B5_{1}$ \n(for chance at least 1 crossing > 0.1) \nAlternative mask", cbar_kws={'label': 'ratio [-]','shrink':0.4}, cmap = cmap, mask = mask_cross, n_xticks=8, n_yticks=6)
# im.plot_heatmap(first_crossings_CI, title = "Indium Selenide Sample \nrelative broadness CI energy first crossings $\u03B5_{1}$ \n(for chance at least 1 crossing > 0.1), capped at 0.005 \nAlternative mask", cbar_kws={'label': 'ratio [-]','shrink':0.4}, vmax = 0.005, cmap = cmap, mask = mask_cross, n_xticks=8, n_yticks=6)


#%% ENERGY CROSSINGS ??

# first_crossings = np.zeros(np.append(im.image_shape, 3))
# first_crossings_CI = np.zeros(im.image_shape)
# for i in range(im.image_shape[0]):
#     for j in range(im.image_shape[1]):
#         if im.n_cross[i,j,0] > 1:
#             if len(im.n_cross[i,j]) > 0:
#                 first_crossings[i,j,:] = im.n_cross[i,j][0,:]
#                 first_crossings_CI[i,j] = im.n_cross[i,j][0,2]-im.n_cross[i,j][0,1]
#
# im.plot_heatmap(first_crossings[:,:,0], title = "energy first crossing real part dielectric function\n(for chance at least 1 crossing > 0.5)", cbar_kws={'label':  'energy [eV]'}, cmap = cmap)
# im.plot_heatmap(first_crossings_CI, title = "broadness CI energy first crossing real part dielectric function \n(for chance at least 1 crossing > 0.5)", cbar_kws={'label':  'energy [eV]'}, cmap = cmap)


#%% ENERGY CROSSING DISCRETIZED

mask_E_cross = (mask | (im.n_cross[:,:,0] == 0))
size_E_cross_bins = 0.5 #round_to_nearest(np.nanpercentile((first_crossings[:,:,2]-first_crossings[:,:,1])[~mask_E_cross],50)/0.1,1.0)
E_cross_round  = np.round(first_crossings[:,:,0]/size_E_cross_bins) * size_E_cross_bins
im.plot_heatmap(E_cross_round, title = title_specimen + r"$\rm{-\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale}, color_bin_size = size_E_cross_bins, discrete_colormap = True, sig_cbar = 2,
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross, vmin = 21, vmax = 25,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_Crossings_Discretized')

size_E_cross_bins = 0.5 #round_to_nearest(np.nanpercentile((first_crossings[:,:,2]-first_crossings[:,:,1])[~mask_E_cross],50)/0.5,0.5)
E_cross_round  = np.round(first_crossings[:,:,0]/size_E_cross_bins) * size_E_cross_bins
im.plot_heatmap(E_cross_round, title = title_specimen + r"$\rm{-\;Energy\;Max\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale}, color_bin_size = size_E_cross_bins, discrete_colormap = True, sig_cbar = 2,
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_max_cross,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_Max_Crossings_Discretized')

size_E_cross_error_bins = 0.001 #round_to_nearest(np.nanpercentile((first_crossings[:,:,2]-first_crossings[:,:,1])[~mask_E_cross],50)/0.1,1.0)
E_cross_error_round  = np.round(first_crossings_CI[:,:]/size_E_cross_error_bins) * size_E_cross_error_bins
im.plot_heatmap(E_cross_error_round, title = title_specimen + r"$\rm{-\;Relative\;Error\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, color_bin_size = size_E_cross_error_bins, discrete_colormap = True, sig_cbar = 2,
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_cross, vmax = 0.009,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_First_Crossings_Error_discretized')

size_E_cross_error_bins = 0.001 #round_to_nearest(np.nanpercentile((first_crossings[:,:,2]-first_crossings[:,:,1])[~mask_E_cross],50)/0.1,1.0)
E_cross_error_round  = np.round(first_crossings_CI[:,:]/size_E_cross_error_bins) * size_E_cross_error_bins
im.plot_heatmap(E_cross_error_round, title = title_specimen + r"$\rm{-\;Relative\;Error\;Energy\;First\;Crossings\;}$" + "$\epsilon_{1}$", 
                cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, color_bin_size = size_E_cross_error_bins, discrete_colormap = True, sig_cbar = 2,
                xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
                mask = mask_max_cross, vmax = 0.009,
                sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int, 
                save_as = save_loc + save_title_specimen + '_Energy_Max_Crossings_Error_discretized')

# mask_E_cross = (mask | (im.n_cross[:,:,2] == 0))
# size_E_cross_bins = np.nanpercentile((first_crossings[:,:,2]-first_crossings[:,:,1])[~mask_cross],50)/0.05
# E_cross_round  = np.round(first_crossings[:,:,0]/size_E_bins) * size_E_bins
# im.plot_heatmap(E_cross_round, title = "Indium Selenide Sample \nenergy first crossing $\u03B5_{1}$ discretized \n(for chance at least 1 crossing > 0.1) \nAlternative mask", cbar_kws={'label': 'energy [eV]','shrink':0.4}, cmap = cmap, mask = mask_E_cross, color_bin_size = size_E_cross_bins, discrete_colormap = True, sig=3, n_xticks=8, n_yticks=6)

#%% DIELECTRIC FUNCTION INDIVIDUAL PIXELS

for i in np.arange(0, 31, 30): # len(im.x_axis)
    for j in np.arange(0, 31, 30): #len(im.y_axis)
        if i != 0 and j != 0:
            pixx=20
            pixy=80
            
            epsilon1 = im.eps[pixy,pixx,0].real
            epsilon2 = im.eps[pixy,pixx,0].imag
            
            fig1, ax1 = plt.subplots(dpi=200)
            ax1.plot(im.deltaE[(len(im.deltaE)-len(epsilon1)):], epsilon1, label = "$\epsilon_{1}$")
            ax1.plot(im.deltaE[(len(im.deltaE)-len(epsilon2)):], epsilon2, label = "$\epsilon_{2}$")
            ax1.axhline(0, color='black')
            ax1.set_title(title_specimen + r"$\rm{-\;Dielectric\;Function\;pixel[%d,%d]}$"%(pixx, pixy))
            ax1.set_xlabel(r"$\rm{Energy\;Loss\;[eV]\;}$")
            ax1.set_ylabel(r"$\rm{Dielectric\;Function\;[F/m]\;}$")
            ax1.set_ylim(-0.2,5)
            ax1.legend()

            plt.savefig(save_loc + save_title_specimen + '_Dielectric_function_pixel[' + str(pixx) + ','+ str(pixy) + '].pdf')
        
#%% BANDGAP

# im.plot_heatmap(im.E_band[:,:,0], title = title_specimen + r"$\rm{-\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_Fixed_b')

# im.plot_heatmap(im.E_band[:,:,0], title = title_specimen + r"$\rm{-\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 1.5,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_Fixed_b_capped')

# im.plot_heatmap((im.E_band[:,:,2]-im.E_band[:,:,1])/(2*im.E_band[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_Error_Fixed_b')

# im.plot_heatmap((im.E_band[:,:,2]-im.E_band[:,:,1])/(2*im.E_band[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 0.3,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_Error_Fixed_b_capped')


# im.plot_heatmap((im.E_band[:,:,2]-im.E_band[:,:,1])/(im.E_band[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Broadness\;CI\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_CI')

# im.plot_heatmap((im.E_band[:,:,2]-im.E_band[:,:,1])/(im.E_band[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Broadness\;CI\;Bandgap\;Energy\;}$" + "\n" + r"$\rm{Capped\;at\;0.1\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 0.1,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_CI_capped')

#%% BANDGAP EXPONENT

# im.plot_heatmap(im.b[:,:,0], title = title_specimen + r"$\rm{-\;Bandgap\;Exponent\;}$",
#                 cbar_kws={'label': r"$\rm{[-]\;}$", 'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponent')

# im.plot_heatmap(im.b[:,:,0], title = title_specimen + r"$\rm{-\;Bandgap\;Exponent\;}$" + "\n" + r"$\rm{b\;<\;1\;}$",
#                 cbar_kws={'label': r"$\rm{[-]\;}$", 'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmin=0, vmax=1,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponent_capped')

# im.plot_heatmap((im.b[:,:,2]-im.b[:,:,1])/(2*im.b[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Bandgap\;Exponent\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponenent_Error')

# im.plot_heatmap((im.b[:,:,2]-im.b[:,:,1])/(2*im.b[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Error\;Bandgap\;Exponent\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 1.0,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponenent_Error_capped')

# im.plot_heatmap((im.b[:,:,2]-im.b[:,:,1])/(im.b[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Broadness\;CI\;Bandgap\;Exponent\;}$",
#                 cbar_kws={'label': 'Ratio [-]','shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponenent_CI')

# im.plot_heatmap((im.b[:,:,2]-im.b[:,:,1])/(im.b[:,:,0]), title = title_specimen + r"$\rm{-\;Relative\;Broadness\;CI\;Bandgap\;Exponent\;}$" + "\n" + r"$\rm{Capped\;at\;0.2\;}$",
#                 cbar_kws={'label': 'Ratio [-] ','shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 0.2,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponent_CI_capped')

#%% BANDGAP DISCRETIZED

# mask_E_band = (mask | ((im.E_band[:,:,2]-im.E_band[:,:,1])/im.E_band[:,:,0] >= 1))
# size_E_band_bins = 0.1 #round_to_nearest(np.nanpercentile((im.E_band[:,:,2]-im.E_band[:,:,1])[~mask_E_band],50)/8,0.5)
# E_band_round  = np.round(im.E_band[:,:,0]/size_E_band_bins) * size_E_band_bins
# im.plot_heatmap(E_band_round, title = title_specimen + r"$\rm{-\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Energy\;[eV]\;}$",'shrink':cb_scale}, color_bin_size = size_E_band_bins, discrete_colormap = True, sig_cbar = 2,
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 1.4,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_Discretized_Fixed_b')

# E_band_error = (im.E_band[:,:,2]-im.E_band[:,:,1])/(2*im.E_band[:,:,0])
# size_E_band_error_bins = 0.05 #round_to_nearest(np.nanpercentile((im.E_band[:,:,2]-im.E_band[:,:,1])[~mask_E_band],50)/8,0.5)
# E_band_error_round  = np.round(E_band_error/size_E_band_error_bins) * size_E_band_error_bins
# im.plot_heatmap(E_band_error_round, title = title_specimen + r"$\rm{-\;Relative\;Error\;Bandgap\;Energy\;}$",
#                 cbar_kws={'label': r"$\rm{Ratio\;[-]\;}$",'shrink':cb_scale}, color_bin_size = size_E_band_error_bins, discrete_colormap = True, sig_cbar = 3,
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_Relative_Error_Discretized_Fixed_b')

#%% BANDGAP EXPONENT

# mask_b = (mask | (im.b[:,:,0] == 0))
# size_b_bins = 0.2 #round_to_nearest(np.nanpercentile((im.b[:,:,2]-im.b[:,:,1])[~mask_b],50)/8,0.2)
# b_round  = np.round(im.b[:,:,0]/size_b_bins) * size_b_bins
# im.plot_heatmap(b_round, title = title_specimen + r"$\rm{-\;Bandgap\;Exponent\;}$",
#                 cbar_kws={'label':'[-]','shrink':cb_scale}, color_bin_size = size_b_bins, discrete_colormap = True, sig_cbar = 2,
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 1,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponent_Discretized')

# b_error = (im.b[:,:,2]-im.b[:,:,1])/(2*im.b[:,:,0])
# size_b_error_bins = 0.2 #round_to_nearest(np.nanpercentile((im.b[:,:,2]-im.b[:,:,1])[~mask_b],50)/8,0.2)
# b_error_round  = np.round(b_error/size_b_error_bins) * size_b_error_bins
# im.plot_heatmap(b_error_round, title = title_specimen + r"$\rm{-\;Relative\;Error\;Bandgap\;Exponent\;}$",
#                 cbar_kws={'label':'[-]','shrink':cb_scale}, color_bin_size = size_b_bins, discrete_colormap = True, sig_cbar = 2,
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask, vmax = 1,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_exponent_Discretized')

#%% MSE

# mse_max = np.max(im.mse[:,:,0])
# mse_norm = im.mse[:,:,0] / mse_max
#
#
# im.plot_heatmap(mse_norm, title = title_specimen + r"$\rm{-\;Bandgap\;MSE\;}$",
#                 cbar_kws={'label': r"$\rm{[-]\;}$",'shrink':cb_scale},
#                 xlab = r"$\rm{[nm]\;}$", ylab = r"$\rm{[nm]\;}$", cmap = cmap,
#                 mask = mask,
#                 sig_ticks = sig_ticks, scale_ticks = scale_ticks, npix_xtick = npix_xtick, npix_ytick = npix_ytick, tick_int = tick_int,
#                 save_as = save_loc + save_title_specimen + '_Bandgap_MSE_Fixed_b')



#%% BANDGAP FIT INDIVIDUAL PIXELS
"""
def bandgap_test(x, amp, BG, b=1.5):
    result = np.zeros(x.shape)
    result[x<BG] = 0
    result[x>=BG] = amp * (x[x>=BG] - BG)**(b)
    return result

for i in np.arange(0, im.image_shape[1], 30):
    for j in np.arange(0, im.image_shape[0], 30):
        if i != 0 and j != 0:
            pixx=i
            pixy=j
            windowlength = 21
            polyorder = 2
            #dE1 = im.dE1[im.clustered]
            
            ieels_p = im.ieels_p[pixy,pixx,0,:]
            ieels_p_smooth = savgol_filter(ieels_p, window_length = windowlength, polyorder = polyorder)
            ieels_p_1d = np.diff(ieels_p_smooth)
            ieels_p_1d_smooth = savgol_filter(ieels_p_1d, window_length = windowlength, polyorder = polyorder)
            
            try:
                range1 = im.deltaE[ieels_p_smooth > 2]
                range1 = range1[range1 > 0][0]
                
                range3 = im.deltaE[np.argwhere(np.diff(np.sign(ieels_p_1d_smooth + 1)))]
                range3 = range3[range3 > range1][0]
                
                range2 = (range3+range1)/2
                #baseline = np.average(IEELSs_fit[(im.deltaE > range1 - 0.1) & (im.deltaE < range1)])
                
                popt, pcov = curve_fit(bandgap_test, im.deltaE[(im.deltaE > range1) & (im.deltaE < range2)], 
                                       ieels_p[(im.deltaE > range1) & (im.deltaE < range2)],
                                       p0 = [300, 1.5], bounds=([0, 0], np.inf))
                
                Amp = popt[0]
                E_bg = popt[1]
                #b = popt[2]
                print("succes! Range1 = " + str(range1) + ", Range2 = " + str(range2))
            except:
                print("frowny face")


            fig2, ax2 = plt.subplots(dpi=200)
            ax2.set_title(title_specimen + r"$\rm{-\;Bandgap\;Fit\;pixel[%d,%d]}$"%(pixx, pixy))
            ax2.set_xlabel(r"$\rm{Energy\;Loss\;[eV]\;}$")
            ax2.set_ylabel(r"$\rm{Intensity\;[a.u.]\;}$")
            ax2.set_ylim(-2,600)
            ax2.set_xlim(1,5)
            
            ax2.fill_between(im.deltaE, im.ieels_p[pixy,pixx,1,:], im.ieels_p[pixy,pixx,2,:], alpha = 0.1, color = 'C0')
            ax2.plot(im.deltaE, ieels_p,label = "Spectrum", alpha = 1.0, color = 'C0')
            ax2.plot(im.deltaE, bandgap_test(im.deltaE, Amp, E_bg), label = r"$\rm{Fit}$", alpha = 1.0, color = 'C1')
            ax2.axhline(0, color = 'black', alpha=0.5)
            ax2.axvspan(xmin=range1, xmax=range2, ymin=-1000, ymax=1000, color = 'C3', alpha=0.1)
            ax2.legend(loc="lower right",frameon=False)

            #plt.savefig(save_loc + save_title_specimen + '_Bandgap_fit_pixel[' + str(pixx) + ','+ str(pixy) + '].pdf')
            
            # Free b
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt[1],4)) + ", b = " + str(round(popt[2],4)))
            
            # Fixed b
            print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, BG = " + str(round(popt[1],4)))
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt[1],4)))
"""
#%% BANDGAP FIT INDIVIDUAL PIXELS ALL MODELS
"""
def bandgap(x, amp, BG, b = 1.5):
    global baseline
    result = np.zeros(x.shape)
    result[x<BG] = 0
    result[x>=BG] = amp * (x[x>=BG] - BG)**(b)
    return result

#path_to_image = 'C:/Users/abelbrokkelkam/PhD/data/dmfiles/10n-dop-inse-B1_stem-eels-SI-processed_003.dm4'
path_to_image = 'C:/Users/abelbrokkelkam/PhD/data/dmfiles/area03-eels-SI-aligned.dm4'
im = SpectralImage.load_data(path_to_image)

#path_to_models = 'C:/Users/abelbrokkelkam/PhD/data/MLdata/models/dE_n10-inse_SI-003/E1_shift09_k10_average/'
path_to_models = 'C:/Users/abelbrokkelkam/PhD/data/MLdata/models/dE_nf-ws2_SI-001/E1_p16_k5_average/'
im.load_zlp_models(path_to_models=path_to_models)
im.cluster(5)
im.pool(5)
im.calc_axes()

im.e0 = 200									# keV
im.beta = 21.3								# mrad
im.set_n(3.0)								# refractive index, InSe no background
#%%
for i in np.arange(0, 31, 30): # im.image_shape[1]
    for j in np.arange(0, 31, 30): # im.image_shape[0]
        if i != 0 and j != 0:
            pixx=20
            pixy=80
            cluster = im.clustered[pixy,pixx]
            dE1 = im.dE1[1,int(cluster)]
            dE2 = dE1 * 3
            
            #[ts, IEELSs, max_IEELSs], [epss, ts_p, S_ss_p, IEELSs_p, max_IEELSs_p] = im.KK_pixel(pixy, pixx, signal="pooled", iterations=5, select_ZLPs=False)
            
            signal_bg = im.get_pixel_signal(pixy, pixx, signal="pooled")
            ZLPs_match_bg = im.calc_zlps_matched(pixy, pixx, signal="pooled", select_ZLPs=False)
            IEELSs_p_bg = signal_bg - ZLPs_match_bg
            
            hiccup_check = np.max(signal_bg[(im.deltaE > dE1) & (im.deltaE < dE2)])
            hiccup_idx = np.argmax(signal_bg[(im.deltaE > dE1) & (im.deltaE < dE2)])
            idx_hiccup = None
            if hiccup_check > 2000:
                idx_hiccup = np.argmax(signal_bg[(im.deltaE > dE1) & (im.deltaE < dE2)]) + len(signal_bg[im.deltaE <= dE1])
            
            #%%
            n_model         = IEELSs_p_bg.shape[0]
            windowlength    = 21
            polyorder       = 2
            
            IEELSs_p_bg_smooth     = savgol_filter(IEELSs_p_bg, window_length = windowlength, polyorder = polyorder, axis = 1)
            IEELSs_p_bg_1d         = np.diff(IEELSs_p_bg_smooth, axis = 1)
            IEELSs_p_bg_1d_smooth  = savgol_filter(IEELSs_p_bg_1d, window_length = windowlength, polyorder = polyorder, axis = 1)
            #IEELSs_p_2d = np.diff(IEELSs_p_1d_smooth, axis = 1)
            #IEELSs_p_2d_smooth = savgol_filter(IEELSs_p_2d, window_length = windowlength, polyorder = polyorder, axis = 1)
            
            As = []
            E_bands = []
            bs = []
            bandgapfits = []
            
            k_succes = []
            range1list = []
            range2list = []
            
            mselist = []
            for k in range(n_model):
                IEELSs_fit = IEELSs_p_bg[k]
                IEELSs_fit_smooth = IEELSs_p_bg_smooth[k]
                
                try:
                    range1 = dE1 * 0.85
                    #range1 = 0.7*dE1
                    
                    #range1 = im.deltaE[IEELSs_fit_smooth > 15]
                    #range1 = range1[range1 > 0][0]
                    
                    #range3 = im.deltaE[np.argwhere(np.diff(np.sign(IEELSs_p_1d_smooth[k] + 1)))]
                    #range3 = range3[range3 > range1]
                    #range3 = range3[range3 > dE1][0]
                    
                    #range2 = (range3+range1)/2
                    
                    range2 = range1 + 1.0
                    
                    deltaE_filter = (im.deltaE > range1) & (im.deltaE < range2)
                    
                    if idx_hiccup is not None:
                        hiccup_filter = (im.deltaE > im.deltaE[idx_hiccup - 5]) & (im.deltaE < im.deltaE[idx_hiccup + 5])
                        
                        for i in range(len(deltaE_filter)):
                            if deltaE_filter[i] == True and hiccup_filter[i] == True:
                                deltaE_filter[i] = False
                    
                    fit_domain_x = im.deltaE[deltaE_filter]
                    fit_domain_y = IEELSs_fit[deltaE_filter]
                    #baseline = np.average(IEELSs_fit[(im.deltaE > range1 - 0.1) & (im.deltaE < range1)])
                    baseline = np.average(IEELSs_fit[(im.deltaE < range1)])
                    
                    popt, pcov = curve_fit(bandgap, fit_domain_x, fit_domain_y, 
                                           p0 = [IEELSs_fit[im.deltaE >= range2][0], range1], 
                                           bounds=([0, 0], np.inf))
                    
                    As.append(popt[0])
                    E_bands.append(popt[1])
                    #bs.append(popt[2])
                    
                    bandgapfits.append(bandgap(im.deltaE, popt[0], popt[1]))
                    #bandgapfits.append(bandgap_test(im.deltaE, popt[0], popt[1], popt[2]))
                    
                    k_succes.append(k)
                    range1list.append(range1)
                    range2list.append(range2)
                    
                    mse_value = np.mean(np.square(bandgap(fit_domain_x, *popt) - fit_domain_y))
                    mselist.append(mse_value)
                    print("succes!")
                except:
                    #n_fails += 1
                    #print("fail nr.: ", n_fails, "failed curve-fit, row: ", row, ", pixel: ", j, ", model: ", i)
                    print("frowny face")
            
            As = np.array(As)
            E_bands = np.array(E_bands)
            bs = np.array(bs)
            bandgapfits = np.array(bandgapfits)
            range1list = np.array(range1list)
            range2list = np.array(range2list)
            k_succes = np.array(k_succes)
            mselist = np.array(mselist)
            #%%
            #title_specimen = r'$\rm{InSe\;}$'
            #save_title_specimen = 'InSe'
            #save_loc = "C:/Users/abelbrokkelkam/PhD/data/MLdata/plots/paper/InSe/"
            
            title_specimen = r'$\rm{WS_2\;nanoflower\;}$'
            save_title_specimen = 'WS2_nanoflower_flake'
            save_loc = "C:/Users/abelbrokkelkam/PhD/data/MLdata/plots/paper/WS2/"
            import random
            
            
            fig2, ax2 = plt.subplots(dpi=200)
            ax2.set_title(title_specimen + r"$\rm{-\;Bandgap\;Fit\;pixel[%d,%d]}$"%(pixx, pixy))
            ax2.set_xlabel(r"$\rm{Energy\;Loss\;[eV]\;}$")
            ax2.set_ylabel(r"$\rm{Intensity\;[a.u.]\;}$")
            ax2.set_ylim(-2,500)
            ax2.set_xlim(1,5)
            
            ax2.fill_between(im.deltaE, np.nanpercentile(IEELSs_p_bg_smooth, 16, axis = 0), np.nanpercentile(IEELSs_p_bg_smooth, 84, axis = 0), alpha = 0.1, color = 'C0')
            #ax2.fill_between(im.deltaE[1:], np.nanpercentile(IEELSs_p_1d_smooth, 16, axis = 0), np.nanpercentile(IEELSs_p_1d_smooth, 84, axis = 0), alpha = 0.1, color = 'C1')
            #ax2.fill_between(im.deltaE[1:-1], np.nanpercentile(IEELSs_p_2d_smooth, 16, axis = 0), np.nanpercentile(IEELSs_p_2d_smooth, 84, axis = 0), alpha = 0.1, color = 'C2')
            #ax2.plot(im.deltaE, np.nanpercentile(IEELSs_p_bg, 50, axis = 0),label=r"$\rm{Median\;all\;model}$", alpha = 0.1, color = 'C0')
            ax2.plot(im.deltaE, np.nanpercentile(IEELSs_p_bg_smooth, 50, axis = 0),label=r"$\rm{Median\;substracted\;spectra}$", alpha = 0.5, color = 'C0')
            #ax2.plot(im.deltaE[1:], np.nanpercentile(IEELSs_p_1d_smooth, 50, axis = 0), alpha = 0.5, color = 'C1')
            #ax2.plot(im.deltaE[1:-1], np.nanpercentile(IEELSs_p_2d_smooth, 50, axis = 0), alpha = 0.5, color = 'C2')
            
            
            # Plot individual models
            
            for k in range(1):
                r = random.random()
                g = random.random()
                b = random.random()
                color = 'C1'#(r ,g ,b)
                bg_idx = np.random.randint(0, len(k_succes))
                ax2.plot(im.deltaE, bandgapfits[bg_idx], color=color,label=r"$\rm{Fit\;single\;model}$")
                ax2.plot(im.deltaE, IEELSs_p_bg[k_succes[bg_idx]], color=color, alpha=0.1)
                ax2.plot(im.deltaE, IEELSs_p_bg_smooth[k_succes[bg_idx]], color=color, alpha=0.5)
                ax2.axvspan(xmin=range1list[bg_idx], xmax=range2list[bg_idx], ymin=-1000, ymax=1000, color = color, alpha=0.1)
                #ax2.text(x=0.1,y=350,s=r"$\rm{Amp\;=\;%.2f\;}$"%(As[bg_idx]))
                #ax2.text(x=0.1,y=320,s=r"$\rm{Ebg\;=\;%.2f\;eV}$"%(E_bands[bg_idx]))
                #ax2.text(x=0.1,y=290,s=r"$\rm{b\;=\;%.2f}$"%(bs[bg_idx]))
                #ax2.text(x=0.1,y=290,s=r"$\rm{MSE\;=\;%.2f}$"%(mselist[bg_idx]))
                print(bg_idx)
                print(range1, range2)
            
            ax2.fill_between(im.deltaE, np.nanpercentile(bandgapfits, 16, axis = 0), np.nanpercentile(bandgapfits, 84, axis = 0), alpha = 0.1, color = 'C2')
            # ax2.fill_between(im.deltaE,
            #                 bandgap_test(im.deltaE, np.nanpercentile(As, 16, axis = 0),np.nanpercentile(E_bands, 16, axis = 0),np.nanpercentile(bs, 16, axis = 0)),
            #                 bandgap_test(im.deltaE, np.nanpercentile(As, 84, axis = 0),np.nanpercentile(E_bands, 84, axis = 0),np.nanpercentile(bs, 84, axis = 0)),
            #                 alpha = 0.2, color = 'C5')
            # ax2.plot(im.deltaE, np.nanpercentile(bandgapfits, 50, axis = 0),label = "median bandgapfits", alpha = 1.0, color = 'C4')
            # ax2.plot(im.deltaE, bandgap_test(im.deltaE, np.nanpercentile(As, 50, axis = 0),np.nanpercentile(E_bands, 50, axis = 0),np.nanpercentile(bs, 50, axis = 0)),label = r"$\rm{median\;parameters}$", alpha = 1.0, color = 'C4')
            ax2.plot(im.deltaE, bandgap(im.deltaE, np.nanpercentile(As, 50, axis = 0),np.nanpercentile(E_bands, 50, axis = 0)),label = r"$\rm{Fit\;median\;parameters}$", alpha = 1.0, color = 'C2', linestyle='--')
            #ax2.text(x=0.1,y=250,s=r"$\rm{MSE\;Median\;=\;%.2f}$"%(np.nanpercentile(mselist, 50, axis = 0)))
            ax2.axhline(0,color = 'black', alpha=0.5)
            #ax2.axvspan(xmin=np.nanpercentile(range1, 50, axis = 0), xmax=np.nanpercentile(range2, 50, axis = 0), ymin=-1000, ymax=1000, color = 'C3', alpha=0.1)
            ax2.legend(loc=4,frameon=False)

            fig2.savefig(save_loc + save_title_specimen + '_Bandgap_fit_pixel[' + str(pixx) + ','+ str(pixy) + '].pdf')
            
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt[1],4)))
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt[1],4)) + ", b = " + str(round(popt[2],4)))
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt2[1],4)) + ", b = " + str(round(popt2[2],4)) + " (smooth)")
            
            # Fixed b
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt[1],4)))
            #print("pixel[" + str(pixx) + ","+ str(pixy) + "] done, dE1 = " + str(round(dE1,4)) + ", BG = " + str(round(popt2[1],4)) + " (smooth)")
"""
#%% EPSILON

"""
eps = im.eps[30,30,0,:]
eps_1 = eps.real
eps_2 = eps.imag
deltaE_eps = np.arange(0,len(eps)*15, 15)/1000

#eps_1_DM = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_1_pixel_30_30_m3d038_to_26d422.txt')
#eps_2_DM = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_2_pixel_30_30_m3d038_to_26d422.txt')
eps_1_DM = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_1_pixel_30_30_m3d038_to_26d422_single_iter_reftail.txt')
eps_2_DM = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_2_pixel_30_30_m3d038_to_26d422_single_iter_reftail.txt')

eps_1_DM_n20 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_1_pixel_30_30_m3d038_to_26d422_20_iter_reftail.txt')
eps_2_DM_n20 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_2_pixel_30_30_m3d038_to_26d422_20_iter_reftail.txt')


#eps_1_DM_n1 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_1_pixel_30_30_m3d038_to_26d422_single_iter.txt')
#eps_2_DM_n1 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Epsilon_1_pixel_30_30_m3d038_to_26d422_single_iter.txt')

fig3, ax3 = plt.subplots(dpi=200)
fig4, ax4 = plt.subplots(dpi=200)
ax3.set_title("epsilon 1 pixel[30,30]")
ax3.set_xlabel("energy loss [eV]")
#ax3.set_ylabel("Crossings")
ax4.set_title("epsilon 2 pixel[30,30]")
ax4.set_xlabel("energy loss [eV]")
#ax4.set_ylabel("Crossings")
ax3.set_ylim(-1,7)
ax4.set_ylim(0,6)
#ax3.set_xlim(-3,5)
#ax3.fill_between(im.deltaE, p_low, p_high, alpha = 0.2)
ax3.plot(deltaE_eps, eps_1, label = "EELSfitter")
ax3.plot(im.deltaE, eps_1_DM, label = "DM, n=1")
ax3.plot(im.deltaE, eps_1_DM_n20, label = "DM, n=20")
ax3.axhline(y=0, color='black', linestyle='-')
ax4.plot(deltaE_eps, eps_2, label = "EELSfitter")
ax4.plot(im.deltaE, eps_2_DM, label = "DM, n=1")
ax4.plot(im.deltaE, eps_2_DM_n20, label = "DM, n=20")
ax4.axhline(y=0, color='black', linestyle='-')

ax3.legend()
ax4.legend()
"""
#%% Surface loss function and intensity
"""
Surloss_func_n1 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Surface-loss-function_pixel_30_30_m3d038_to_26d422_single_iter_reftail.txt')
Surloss_func_n20 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Surface-loss-function_pixel_30_30_m3d038_to_26d422_20_iter_reftail.txt')

Surloss_int_n1 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Surface-loss-intensity_pixel_30_30_m3d038_to_26d422_single_iter_reftail.txt')
Surloss_int_n20 = np.loadtxt('C:/Users/abelbrokkelkam/PhD/data/m20210331/eels/eels-SI/10n_dop_inse-B1_stem-eels-SI-processed_003_Surface-loss-intensity_pixel_30_30_m3d038_to_26d422_20_iter_reftail.txt')


fig5, ax5 = plt.subplots(dpi=200)
fig6, ax6 = plt.subplots(dpi=200)
ax5.set_title("Surface-Loss function pixel[30,30]")
ax5.set_xlabel("energy loss [eV]")
#ax5.set_ylabel("Crossings")
ax6.set_title("Surface-Loss intensity pixel[30,30]")
ax6.set_xlabel("energy loss [eV]")
#ax6.set_ylabel("Crossings")
#ax5.set_ylim(-1,7)
#ax6.set_ylim(0,6)
#ax5.set_xlim(-3,5)
#ax5.fill_between(im.deltaE, p_low, p_high, alpha = 0.2)
#ax5.plot(deltaE_eps, eps_1, label = "EELSfitter")
ax5.plot(im.deltaE, Surloss_func_n1, label = "n=1")
ax5.plot(im.deltaE, Surloss_func_n20, label = "n=20")
ax5.axhline(y=0, color='black', linestyle='-')
#ax6.plot(deltaE_eps, eps_2, label = "EELSfitter")
ax6.plot(im.deltaE, Surloss_int_n1, label = "n=1")
ax6.plot(im.deltaE, Surloss_int_n20, label = "n=20")
ax6.axhline(y=0, color='black', linestyle='-')

ax5.legend()
ax6.legend()
"""
#%%




import matplotlib.pyplot as plt
import numpy as np
from matplotlib import rc
import warnings
import seaborn as sns
from matplotlib import cm

from ..core.spectral_image import round_scientific


def plot_heatmap(image, data, title=None, xlab='[nm]', ylab='[nm]', cmap='coolwarm',
                 discrete_colormap=False, sig_cbar=3, color_bin_size=None, equal_axis=True,
                 sig_ticks=2, npix_xtick=10, npix_ytick=10, scale_ticks=1, tick_int=False,
                 save_as=False, **kwargs):
    """
    Plots a heatmap for given data input.

    Parameters
    ----------
    image: SpectralImage
        :py:meth:`spectral_image.SpectralImage <EELSFitter.core.spectral_image.SpectralImage>` object
    data : numpy.ndarray, shape=(M,)
        Input data for heatmap, but be 2 dimensional.
    title : str, optional
        Set the title of the heatmap. The default is None.
    xlab : str, optional
        Set the label of the x-axis. Nanometer ([nm]) is assumed as standard scale. The default is '[nm]'.
    ylab : str, optional
        Set the label of the y-axis. Nanometer ([nm]) is assumed as standard scale. The default is '[nm]'.
    cmap : str, optional
        Set the colormap of the heatmap. The default is 'coolwarm'.
    discrete_colormap : bool, optional
        Enables the heatmap values to be discretised. Best used in conjuction with color_bin_size. The default is False.
    sig_cbar : int, optional
        Set the amount of significant numbers displayed in the colorbar. The default is 3.
    color_bin_size : float, optional
        Set the size of the bins used for discretisation. Best used in conjuction discrete_colormap. The default is None.
    equal_axis : bool, optional
        Enables the pixels to look square or not. The default is True.
    sig_ticks : int, optional
        Set the amount of significant numbers displayed in the ticks. The default is 2.
    npix_xtick : float, optional
        Display a tick per n pixels in the x-axis. Note that this value can be a float. The default is 10.
    npix_ytick : float, optional
        Display a tick per n pixels in the y-axis. Note that this value can be a float. The default is 10.
    scale_ticks : float, optional
        Change the scaling of the numbers displayed in the ticks. Nanometer ([nm]) is assumed as standard scale adjust scaling from there. The default is 1.
    tick_int : bool, optional
        Set whether you only want the ticks to display as integers instead of floats. The default is False.
    save_as : str, optional
        Set the location and name for the heatmap to be saved to. The default is False.
    **kwargs : dictionary
        Additional keyword arguments.

    """

    fig = plt.figure(dpi=200)
    if title is None:
        if image.name is not None:
            plt.title(image.name)
    else:
        plt.title(title)

    if 'mask' in kwargs:
        mask = kwargs['mask']
        if mask.all():
            warnings.warn("Mask all True: no values to plot.")
            return
    else:
        mask = np.zeros(data.shape).astype('bool')

    if equal_axis:
        plt.axis('scaled')

    if discrete_colormap:

        unique_data_points = np.unique(data[~mask])
        if 'vmax' in kwargs:
            if len(unique_data_points[unique_data_points > kwargs['vmax']]) > 0:
                unique_data_points = unique_data_points[unique_data_points <= kwargs['vmax']]

        if 'vmin' in kwargs:
            if len(unique_data_points[unique_data_points < kwargs['vmin']]) > 0:
                unique_data_points = unique_data_points[unique_data_points >= kwargs['vmin']]

        if color_bin_size is None:
            if len(unique_data_points) == 1:
                color_bin_size = 1
            else:
                color_bin_size = np.nanpercentile(unique_data_points[1:] - unique_data_points[:-1], 30)

        n_colors = round((np.nanmax(unique_data_points) - np.nanmin(unique_data_points)) / color_bin_size + 1)
        cmap = cm.get_cmap(cmap, n_colors)
        spacing = color_bin_size / 2
        kwargs['vmax'] = np.max(unique_data_points) + spacing
        kwargs['vmin'] = np.min(unique_data_points) - spacing

    if image.pixelsize is not None:
        ax = sns.heatmap(data, cmap=cmap, **kwargs)
        xticks, yticks, xticks_labels, yticks_labels = image.get_ticks(sig_ticks, npix_xtick, npix_ytick, scale_ticks,
                                                                      tick_int)
        ax.xaxis.set_ticks(xticks)
        ax.yaxis.set_ticks(yticks)
        ax.set_xticklabels(xticks_labels, rotation=0)
        ax.set_yticklabels(yticks_labels)
    else:
        ax = sns.heatmap(data, **kwargs)

    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    colorbar = ax.collections[0].colorbar
    if discrete_colormap:
        if data.dtype == int:
            colorbar.set_ticks(np.unique(data[~mask]))
        else:
            colorbar.set_ticks(np.unique(data[~mask]))
            cbar_ticks_labels = []
            for tick in np.unique(data[~mask]):
                if tick >= 1:
                    cbar_ticks_labels.append(round_scientific(tick, sig_cbar + len(str(abs(int(math.floor(tick)))))))
                else:
                    cbar_ticks_labels.append(round_scientific(tick, sig_cbar))
            colorbar.ax.set_yticklabels(cbar_ticks_labels)

    if 'vmin' in kwargs:
        if np.nanmin(data[~mask]) < kwargs['vmin']:
            cbar_ticks = colorbar.ax.get_yticklabels()
            loc = 0
            if discrete_colormap:
                loc = np.min(np.argwhere(colorbar.ax.get_yticks() >= kwargs['vmin'] + spacing))
            cbar_ticks[loc] = r'$\leq$' + cbar_ticks[loc].get_text()
            colorbar.ax.set_yticklabels(cbar_ticks)

    if 'vmax' in kwargs:
        if np.nanmax(data[~mask]) > kwargs['vmax']:
            cbar_ticks = colorbar.ax.get_yticklabels()
            cbar_ticks_values = colorbar.ax.get_yticks()
            loc = -1
            if discrete_colormap:
                loc = np.max(np.argwhere(cbar_ticks_values <= kwargs['vmax'] - spacing))
            cbar_ticks[loc] = r'$\geq$' + cbar_ticks[loc].get_text()
            colorbar.ax.set_yticklabels(cbar_ticks)

    if save_as:
        if type(save_as) != str:
            if image.name is not None:
                save_as = image.name
        if 'mask' in kwargs:
            save_as += '_masked'
        save_as += '.pdf'
        plt.savefig(save_as, bbox_inches='tight')
    return fig